from typing import Any, Dict, List, Optional, Tuple

from langchain.chains.llm import LLMChain


class ZLLMChain(LLMChain):
    default_stop: Optional[List[str]] = None

    def prep_prompts(
        self, input_list: List[Dict[str, Any]]
    ) -> Tuple[List[str], Optional[List[str]]]:
        prompts, stop = super().prep_prompts(input_list)
        if stop is None:
            stop = self.default_stop
        return prompts, stop



"""Prompt building for GetMultipleOutputsChain."""

from __future__ import annotations

import dataclasses
from typing import Dict, List, Optional

from langchain.prompts.prompt import PromptTemplate
from pydantic import BaseModel, Extra

from zamm.prompts.parser import DictOutputParser

from .config import VariableConfig


class MultipleOutputsPrompter(BaseModel):
    """Handles the prompting logic for GetMultipleOutputsChain."""

    prefix: str
    """The preamble before we ask the LLM to fill in each variable value."""
    variables: List[VariableConfig]
    """Settings for how each variable is to be displayed and processed."""
    output_parser: Optional[DictOutputParser] = None
    """How to parse the output of calling an LLM on this formatted prompt.

    Must be provided if the LLM is completing all the inputs all at once."""

    class Config:
        """Configuration for this pydantic object."""

        extra = Extra.forbid
        arbitrary_types_allowed = True

    def __init__(
        self,
        prefix: str,
        variables: Optional[Dict[str, str]] = None,
        variable_configs: Optional[List[VariableConfig]] = None,
        auto_suffix_variable_display: bool = True,
        output_parser: Optional[DictOutputParser] = None,
    ):
        r"""Prompt template creator for getting multiple outputs from the LLM.

        Args:
            prefix: Prompt that asks for the values of each of the variables the LLM
                    should fill in
            variables: A mapping from the output key of each variable to its display in
                    the prompt. Use this instead of `variable_configs` if you don't care
                    about customizing per-variable prompts.
            variable_configs: Information about each variable we want the value of. Use
                    this instead of `variables` if you want fine-grained control over
                    the display of variables in the prompt.
            auto_suffix_variable_display: If true, for every variable that doesn't
                    already have a suffix, the display values provided for each
                    variable will have a colon and the variable stop appended to the
                    end.

                    For example, if this is true, and your variable display is
                    "Action Input", this will automatically suffix the stop to get
                    "Action Input: \"" for the final prompt.
            output_parser: The output parser that will be called if this is asked to
                    let the LLM complete all the outputs at once.
        """
        if variables is None and variable_configs is None:
            raise ValueError("Please set either variables or variable_configs")
        elif variables and variable_configs:
            raise ValueError("Please set only one of variables or variable_configs")

        if not variable_configs:
            # this assert helps tell pydantic `variables` is assured to have a value at
            # this point
            assert variables is not None, "Please put the ValueError's back in"
            variable_configs = [
                VariableConfig(display=display, output_key=output_key)
                for output_key, display in variables.items()
            ]

        if auto_suffix_variable_display:
            new_configs = []
            for variable in variable_configs:
                if variable.display_suffix is None:
                    new_suffix = f": {variable.stop.strip()}"
                else:
                    new_suffix = variable.display_suffix  # keep the old one
                new_configs.append(
                    dataclasses.replace(variable, display_suffix=new_suffix)
                )
            variable_configs = new_configs

        super().__init__(
            prefix=prefix, variables=variable_configs, output_parser=output_parser
        )

    @property
    def output_keys(self) -> List[str]:
        """Output keys from this whole chain."""
        return [v.output_key for v in self.variables]

    def _add_variable_values_template(
        self, template: str, inputs: List[VariableConfig]
    ) -> str:
        """Add templating for variable values up until variable i."""
        for input in inputs:
            template += input.prompt_with_value
            if input.stop != "\n":
                template += "\n"
        return template

    def prompt_template_for_variable_at(self, i: int) -> PromptTemplate:
        """Prompt for a single output variable to be filled in."""
        output = self.variables[i]
        inputs = self.variables[:i]
        input_keys = [v.output_key for v in inputs]

        template = self._add_variable_values_template(self.prefix, inputs)
        template += output.prompt

        return PromptTemplate(
            input_variables=input_keys,
            template=template,
        )

    def prompt_template_for_full_input(self) -> PromptTemplate:
        """Prompt for all outputs to be filled in in a single step."""
        if not self.output_parser:
            raise ValueError(
                "Can't prompt for full input if we don't know how to parse it!"
            )

        first_template = self.prompt_template_for_variable_at(0)
        first_template.output_parser = self.output_parser
        return first_template

    def log(self, completions: Dict[str, str]) -> str:
        """Output the entirety of the LLM's inputs in this chain."""
        initial_prompt = self.prompt_template_for_variable_at(0).format()

        final_prompt = self._add_variable_values_template(self.prefix, self.variables)
        final_prompt = final_prompt.format(**completions)

        assert final_prompt.startswith(initial_prompt), "Prompt reconstruction failed"

        return final_prompt[len(initial_prompt) :].strip()



"""Chain that gets multiple outputs from the LLM."""

from typing import Any, Dict, List, Optional

from langchain.chains import SequentialChain
from langchain.chains.base import Chain
from langchain.llms.base import BaseLLM
from pydantic import BaseModel, Extra

from zamm.chains.general.llm import ZLLMChain
from zamm.prompts.parser import DictOutputParser

from .config import VariableConfig
from .prompt import MultipleOutputsPrompter


class GetDictChain(Chain, BaseModel):
    """Chain that gets multiple outputs from the LLM, one at a time by default.

    Use this over manually parsing the outputs if the outputs can be hard to parse
    (e.g. if you're looking for code blocks of indeterminate length and content, it can
    be hard to write a regex to figure out when one block ends and the next one
    starts). It can also be used when the LLM is having difficulty generating all
    outputs in one go. The `one_step` setting lets you easily toggle between one-step
    and multi-step generation for comparison.
    """

    prompter: MultipleOutputsPrompter  #: :meta private:
    # putting these two separately instead of a Union type because pydantic validation
    # fails when the LLMChain does not have values["chains"]
    one_step_chain: Optional[ZLLMChain]  #: :meta private:
    multi_step_chain: Optional[SequentialChain]  #: :meta private:
    completions: Optional[Dict[str, str]]  #: :meta private:

    class Config:
        """Configuration for this pydantic object."""

        extra = Extra.forbid

    def __init__(
        self,
        llm: BaseLLM,
        prefix: str,
        variables: Optional[Dict[str, str]] = None,
        variable_configs: Optional[List[VariableConfig]] = None,
        auto_suffix_variable_display: bool = True,
        output_parser: Optional[DictOutputParser] = None,
        one_step: bool = False,
        one_step_stop: Optional[str] = None,
        **chain_options: Any
    ):
        """Construct a chain that gets multiple outputs from the LLM.

        By default, this chain gets the inputs one at a time, although that can be
        toggled with the `one_step` variable.
        """
        prompter = MultipleOutputsPrompter(
            prefix=prefix,
            variables=variables,
            variable_configs=variable_configs,
            auto_suffix_variable_display=auto_suffix_variable_display,
            output_parser=output_parser,
        )

        one_step_chain = None
        multi_step_chain = None
        if one_step:
            one_step_chain = ZLLMChain(
                llm=llm,
                prompt=prompter.prompt_template_for_full_input(),
                # custom stop when running whole interaction
                default_stop=[one_step_stop],
                **chain_options,
            )
        else:
            chains = []
            for i, var in enumerate(prompter.variables):
                chains.append(
                    ZLLMChain(
                        llm=llm,
                        prompt=prompter.prompt_template_for_variable_at(i),
                        output_key=var.output_key,
                        default_stop=[var.stop],
                        **chain_options,
                    )
                )
            multi_step_chain = SequentialChain(
                chains=chains,
                input_variables=[],
                output_variables=prompter.output_keys,
                **chain_options,
            )

        super().__init__(
            prompter=prompter,
            one_step_chain=one_step_chain,
            multi_step_chain=multi_step_chain,
            **chain_options,
        )

    @property
    def input_keys(self) -> List[str]:
        """Input keys for this chain.

        Always empty because the prefix should have already been pre-formatted before
        being passed into this chain.
        """
        return []

    @property
    def output_keys(self) -> List[str]:
        """Output keys produced by this whole chain."""
        return self.prompter.output_keys

    def log(self) -> str:
        """Return entire transcript for how the LLM filled in these values."""
        assert self.completions, "Run the chain first so we have something to log"
        return self.prompter.log(self.completions)

    def _call(self, inputs: Dict[str, str]) -> Dict[str, str]:
        if self.one_step_chain:
            result = self.one_step_chain.apply_and_parse([inputs])[0]
            # typechecking above should ensure a DictOutputParser gets passed in if
            # we're doing this in one step
            assert isinstance(result, dict), "Please set DictOutputParser"
            self.completions = result
        else:
            assert self.multi_step_chain is not None, "No chains configured"
            self.completions = self.multi_step_chain._call(inputs)

        return self.completions



from langchain.prompts.prompt import PromptTemplate

FOLLOW_TUTORIAL_PROMPT = PromptTemplate(
    input_variables=[],
    template="""
You decide to follow instructions in another file. You must enter in the path to the file (do not pick a different file -- and do not use the link name, but the actual path to the link), and the task you expect to accomplish with this file. For example, if you were to encounter the following step in the training manual:

> Follow the instructions at [`goober.md`](./path/to/goober.md) to floof the goober

then you should enter something like:

Path to the instructions file: `./path/to/goober.md`
Task: Floof the goober

Do it now for the current task.

""".lstrip(),  # noqa
)

FOLLOW_TUTORIAL_LOGGER = PromptTemplate(
    input_variables=["documentation_path", "task"],
    template="\nYou followed the instructions at `{documentation_path}` to '{task}'.",
)



from abc import abstractmethod
from typing import Dict

from langchain.prompts.base import BaseOutputParser


class DictOutputParser(BaseOutputParser):
    """Class to parse the output of an LLM call to a dict."""

    @abstractmethod
    def parse(self, text: str) -> Dict[str, str]:
        """Parse the output of an LLM call."""



"""Configuration for variables in GetMultipleOutputsChain."""

from __future__ import annotations

from dataclasses import dataclass
from typing import Optional, Type, TypeVar

VC = TypeVar("VC", bound="VariableConfig")


@dataclass
class VariableConfig:
    """Configuration for each variable the LLM should fill in."""

    display: str
    """The string that the LLM sees as an invitation to fill this variable in.

    For example, if the value of this is "Action Input: \"", then the LLM will see the
    prompt

    ```
    Decide on your next course of action:

    Action Type: "Search"
    Action Input: "
    ```
    """
    output_key: str
    """The key in the final output dict that the LLM-provided value will be saved to.

    For example, if the value of this is "input", and the LLM prompt completion is:

    ```
    Decide on your next course of action:

    Action Type: "Search"
    Action Input: "Olivia Wilde boyfriend"
    ```

    Then the final dict returned will store "Olivia Wilde boyfriend" inside the key
    "input":

    ```
    {
        ...
        "input": "Olivia Wilde boyfriend"
        ...
    }
    ```
    """
    stop: str = "\n"
    """The character to stop at when completing the value for this variable.

    This defaults to a single newline, so that if your prompt goes

        Action Input:

    then the LLM can know to terminate its input with a newline. If you expect the
    output to be multiline, then you may instead want to go with ''' or ```, and adjust
    the display string accordingly

    This exists here instead of MultipleOutputsPrompter so that each variable can have
    its own custom stop if needed.
    """
    display_suffix: Optional[str] = None
    """The string to put at the end of the display string.

    For example, if you set this to ": \"", and your display is "Action Input", then
    the whole thing will come out as "Action Input: \"" for the LLM to complete

    This exists to allow easy modification of the final display string by simply
    copying over the stopper, without polluting the display string with the stopper.
    """

    @property
    def prompt(self) -> str:
        """Prompt for the LLM to fill in this specific variable."""
        if self.display_suffix:
            return self.display + self.display_suffix
        return self.display

    @property
    def prompt_with_value(self) -> str:
        """Previous prompt + filled-in value for this variable.

        This is a templating string, so it won't have the actual value of the variable,
        just a marker for the template to eventually fill it in with.
        """
        return self.prompt + "{" + self.output_key + "}" + self.stop

    @classmethod
    def for_string(
        cls: Type[VC],
        output_key: str,
        display: str,
        enclosure: str = '"',
    ) -> VC:
        """Configure a variable that should be enclosed in double quotes.

        Optionally use a different enclosure fo the variable.
        """
        return cls(
            output_key=output_key,
            display=display,
            display_suffix=f": {enclosure}",
            stop=enclosure,
        )

    @classmethod
    def for_code(
        cls: Type[VC],
        output_key: str,
        display: str,
        language: Optional[str] = None,
    ) -> VC:
        """Configure a language-specific code prompt."""
        language_str = "" if language is None else language
        display_suffix = f": ```{language_str}\n"
        return cls(
            output_key=output_key,
            display=display,
            display_suffix=display_suffix,
            stop="```",
        )



from typing import Callable, Dict, List

from langchain.agents.agent import AgentExecutor
from langchain.chains.sequential import SequentialChain
from langchain.llms.base import BaseLLM
from langchain.prompts import PromptTemplate
from pydantic import root_validator

from zamm.actions.base import Action
from zamm.actions.edit_file import EditFile
from zamm.actions.finish import Finish
from zamm.actions.follow_tutorial import FollowTutorial
from zamm.actions.note import MakeNote
from zamm.actions.use_terminal import UseTerminal, ZTerminal
from zamm.chains.general import ActionChain
from zamm.chains.general.choice.base import ChoiceChain
from zamm.chains.general.choice.prompt import ChoicePromptTemplate
from zamm.chains.general.llm import ZLLMChain
from zamm.prompts.chained import ChainedPromptTemplate
from zamm.prompts.prefixed import Prefix


class LaxSequentialChain(SequentialChain):
    @root_validator(pre=True)
    def validate_chains(cls, values: Dict) -> Dict:
        return values

    def _validate_outputs(self, outputs: Dict[str, str]) -> None:
        pass

    def _call(self, inputs: Dict[str, str]) -> Dict[str, str]:
        known_values = inputs.copy()
        for i, chain in enumerate(self.chains):
            outputs = chain(known_values, return_only_outputs=True)
            known_values.update(outputs)
        return known_values


def default_action_chain(
    llm: BaseLLM,
    prefix: Prefix,
    terminal: ZTerminal,
    agent_creator: Callable[[], AgentExecutor],
    choice_prompt: str = "You have a few actions available to accomplish this: ",
):
    thought_chain_prompt = ChainedPromptTemplate(
        "",
        prefix,
        PromptTemplate(
            input_variables=[],
            template="""
Now, the next step in the employee training manual is (quoted below as a single line):

> """.lstrip(),
        ),
    )
    thought_chain = ZLLMChain(
        llm=llm,
        output_key="next_step",
        prompt=thought_chain_prompt,
        default_stop=["\n"],
    )

    final_prefix = ChainedPromptTemplate(
        "",
        prefix,
        PromptTemplate(
            input_variables=["next_step"],
            template="""
Now, the next step in the employee training manual is (quoted below as a single line):

> {next_step}

""".lstrip(),
        ),
    )

    actions: List[Action] = [
        MakeNote.default(llm=llm, prefix=final_prefix),
        UseTerminal.default(llm=llm, prefix=final_prefix, terminal=terminal),
        EditFile.default(llm=llm, prefix=final_prefix),
        FollowTutorial.default(
            llm=llm, prefix=final_prefix, agent_creator=agent_creator
        ),
        Finish.default(),
    ]

    action_choice_template = ChoicePromptTemplate(
        prefix=ChainedPromptTemplate("", final_prefix, choice_prompt),
        choices=[action.choice_text for action in actions],
    )
    action_choice = ChoiceChain(
        llm=llm,
        prompt=action_choice_template,
        choice_num_key="action_num",
        choice_key="action",
    )
    action_chain = ActionChain(option_picker=action_choice, actions=actions)
    outputs = action_chain.output_keys
    # todo: clean up chain input/outputs
    outputs.remove("next_step")
    outputs.remove("agent_scratchpad")
    outputs.remove("task")
    outputs.remove("documentation")

    return LaxSequentialChain(
        input_variables=thought_chain.input_keys,
        output_variables=outputs,
        chains=[
            thought_chain,
            action_chain,
        ],
    )



from typing import Any, Callable, Dict

from langchain.agents.agent import AgentExecutor
from langchain.llms.base import BaseLLM
from langchain.schema import AgentAction

from zamm.actions.base import Action
from zamm.agents.z_step import ZStepOutput
from zamm.prompts.chained import ChainedPromptTemplate
from zamm.prompts.prefixed import Prefix

from .chain import FollowTutorialChain
from .prompt import FOLLOW_TUTORIAL_LOGGER, FOLLOW_TUTORIAL_PROMPT


class FollowTutorialOutput(ZStepOutput):
    documentation_path: str
    documentation: str
    task: str
    output: str

    @classmethod
    def from_chain_output(cls, output: Dict[str, Any]):
        return cls(
            decision=AgentAction(
                tool=output["action"],
                tool_input="dummy input",
                log="dummy log",
            ),
            observation="dummy observation",
            documentation_path=output["documentation_path"],
            documentation=output["documentation"],
            task=output["task"],
            output=output["output"],
            logger_template=FOLLOW_TUTORIAL_LOGGER,
        )

    @property
    def template_args(self) -> Dict[str, str]:
        """Construct the dict used to render this output"""
        return {"documentation_path": self.documentation_path, "task": self.task}


class FollowTutorial(Action):
    @classmethod
    def default(
        cls, llm: BaseLLM, prefix: Prefix, agent_creator: Callable[[], AgentExecutor]
    ):
        return cls(
            name="Follow instructions from a file",
            output_type=FollowTutorialOutput,
            chain=FollowTutorialChain(
                llm=llm,
                prompt=ChainedPromptTemplate("", prefix, FOLLOW_TUTORIAL_PROMPT),
                agent_creator=agent_creator,
            ),
        )



"""Fake parsers for testing purposes."""


from typing import Dict, Mapping, Optional

from zamm.prompts.parser import DictOutputParser


class FakeDictParser(DictOutputParser):
    """Fake dict parser for testing purposes."""

    parses: Optional[Mapping] = None

    def parse(self, text: str) -> Dict[str, str]:
        """Return fake parse from input."""
        if self.parses is not None:
            return self.parses[text]
        return {
            "tool": "Fake tool",
            "tool_input": "Fake input",
        }



"""Fake LLM wrapper for testing purposes."""
from typing import Any, List, Mapping, Optional

from langchain.llms.base import LLM
from pydantic import BaseModel


class FakeLLM(LLM, BaseModel):
    """Fake LLM wrapper for testing purposes."""

    queries: Optional[Mapping] = None
    sequenced_responses: Optional[List[str]] = None
    """List of responses to return in order.
    Useful if the prompt is too complicated to reconstruct for testing."""
    num_calls: int = 0
    """Keeps track of which sequenced response to return."""
    ensure_and_remove_stop: bool = False
    """Set to true to check that stops are being set, and being set properly.
    `queries` must be modified to have the stop present.
    """

    @property
    def _llm_type(self) -> str:
        """Return type of llm."""
        return "fake"

    def _check_stop(self, result: str, stop: Optional[List[str]]) -> str:
        if self.ensure_and_remove_stop:
            assert stop is not None, "Stop has not been set"
            found_stop = False
            for s in stop:
                if result.endswith(s):
                    found_stop = True
                    result = result[: len(result) - len(s)]
                    break
            assert found_stop, f"Output '{result}' does not end in {stop}"

        return result

    def _call(self, prompt: str, stop: Optional[List[str]] = None) -> str:
        """First try to lookup in queries, else return 'foo' or 'bar'."""
        self.num_calls += 1

        if self.queries is not None:
            return self._check_stop(self.queries[prompt], stop)
        if self.sequenced_responses is not None:
            return self._check_stop(self.sequenced_responses[self.num_calls - 1], stop)
        if stop is None:
            return "foo"
        else:
            return "bar"

    @property
    def _identifying_params(self) -> Mapping[str, Any]:
        return {}



from typing import Any, Dict, List, Optional

from langchain.agents.agent import AgentExecutor
from langchain.agents.tools import Tool
from langchain.llms.base import BaseLLM
from langchain.schema import AgentAction, AgentFinish

from zamm.actions.use_terminal import ZTerminal
from zamm.chains.bash_action_prompt import (
    EMPLOYEE_TEACHING_INTRO_TEMPLATE,
    FOLLOW_INSTRUCTIONS_TEMPLATE,
)
from zamm.utils import f_join

from .base import CustomAgent
from .employee_actions import default_action_chain
from .memory import AgentMemory, BaseAgentMemory
from .step import StepOutput


class ZammEmployeeBrain(CustomAgent):
    condense_memory: bool = False

    @property
    def _agent_type(self) -> str:
        """Return Identifier of agent type."""
        return "zamm-employee"

    def _get_next_action(self, full_inputs: Dict[str, str]) -> AgentAction:
        raise NotImplementedError()

    def _construct_scratchpad_base(
        self, memory: BaseAgentMemory, condensed: bool
    ) -> str:
        logs = []
        steps = memory.steps()
        for i, output in enumerate(steps):
            previous = steps[i - 1] if i > 0 else None
            next = steps[i + 1] if i < len(steps) - 1 else None
            logs.append(output.log(previous=previous, next=next, condensed=condensed))
        scratch = f_join("\n", logs)
        if scratch == "":
            return scratch
        return "\n" + scratch + "\n"

    def _construct_scratchpad_structured(self, memory: BaseAgentMemory) -> str:
        return self._construct_scratchpad_base(memory, condensed=self.condense_memory)

    def _construct_scratchpad_final(self, memory: BaseAgentMemory) -> str:
        return self._construct_scratchpad_base(memory, condensed=False)


class ZammEmployee(AgentExecutor):
    agent: ZammEmployeeBrain
    terminal: ZTerminal
    max_iterations: Optional[int] = 100

    def __init__(
        self,
        llm: BaseLLM,
        condense_memory: bool = False,
        tools: Optional[List[Tool]] = None,
        terminal_safe_mode: bool = True,
        **kwargs,
    ):
        tools = tools if tools else []
        brain = ZammEmployeeBrain.from_llm(llm)
        brain.condense_memory = condense_memory
        super().__init__(
            agent=brain,
            tools=tools,
            terminal=ZTerminal(safe_mode=terminal_safe_mode),
            **kwargs,
        )

    @property
    def input_keys(self) -> List[str]:
        return ["task", "documentation"]

    def _custom_take_next_step(self, memory: BaseAgentMemory) -> StepOutput:
        if "documentation" in memory.inputs:
            prefix = FOLLOW_INSTRUCTIONS_TEMPLATE
        else:
            prefix = EMPLOYEE_TEACHING_INTRO_TEMPLATE

        def create_new_employee() -> AgentExecutor:
            return self.__class__(
                llm=self.agent.llm_chain.llm,
                condense_memory=self.agent.condense_memory,
                terminal_safe_mode=self.terminal.safe_mode,
            )

        action_chain = default_action_chain(
            llm=self.agent.llm_chain.llm,
            prefix=prefix,
            terminal=self.terminal,
            agent_creator=create_new_employee,
        )
        results = action_chain(
            {
                **memory.inputs,
                "agent_scratchpad": self.agent._construct_scratchpad_structured(memory),
            }
        )
        return results[action_chain.chains[-1].step_output_key]

    def _return_structured(
        self, output: StepOutput, memory: AgentMemory
    ) -> Dict[str, Any]:
        memory.add_step(output)
        task = memory.inputs["task"]
        logs = self.agent._construct_scratchpad_final(memory)
        tutorial = f"""
Say you want to do the following task:

> {task}

You can do so by following these steps:

{logs}
""".strip()
        return {"output": tutorial}

    def _validate_inputs(self, inputs: Dict[str, str]) -> None:
        """Skip validation because some inputs are optional."""

    def _call(self, inputs: Dict[str, str]) -> Dict[str, Any]:
        """Stripped down version of AgentExecutor._call."""
        # Do any preparation necessary when receiving a new input.
        self.agent.prepare_for_new_call()

        memory = AgentMemory(inputs=inputs)
        iterations = 0
        # We now enter the agent loop (until it returns something).
        while self._should_continue(iterations):
            step_result = self._custom_take_next_step(memory)
            # If the tool chosen is the finishing tool, then we end and return.
            if isinstance(step_result.decision, AgentFinish):
                return self._return_structured(step_result, memory)

            memory.add_step(step_result)
            iterations += 1
        output = self.agent.return_stopped_response(
            self.early_stopping_method, memory.as_intermediate_steps(), **inputs
        )
        return self._return(output, memory.as_intermediate_steps())



from typing import Callable, Dict, List

from langchain.agents.agent import AgentExecutor
from langchain.chains.llm import LLMChain
from langchain.prompts.base import BasePromptTemplate
from pydantic import BaseModel

from zamm.chains.general.get_dict import GetDictChain, VariableConfig
from zamm.utils import read_documentation

from .prompt import FOLLOW_TUTORIAL_PROMPT


class FollowTutorialChain(LLMChain, BaseModel):
    """Asks LLM for which tutorial to follow."""

    agent_creator: Callable[[], AgentExecutor]
    prompt: BasePromptTemplate = FOLLOW_TUTORIAL_PROMPT

    @property
    def input_keys(self) -> List[str]:
        """Input keys to this chain."""
        return self.prompt.input_variables

    @property
    def output_keys(self) -> List[str]:
        return ["documentation_path", "documentation", "task", "output"]

    def _call(self, inputs: Dict[str, str]) -> Dict[str, str]:
        prefix = self.prompt.format(**inputs)
        tutorial_dict = GetDictChain(
            llm=self.llm,
            prefix=prefix,
            variable_configs=[
                VariableConfig.for_string(
                    output_key="documentation_path",
                    display="Path to the instructions file",
                    enclosure="`",
                ),
                VariableConfig(
                    output_key="task",
                    display="Task",
                ),
            ],
        )({})
        documentation = read_documentation(tutorial_dict["documentation_path"])

        employee = self.agent_creator()
        args = {
            "documentation": documentation,
            "task": tutorial_dict["task"],
        }
        result = employee(args)
        return {**tutorial_dict, **args, **result}
